import os
import subprocess
import pickle
import gzip
from typing import Dict, List, Union

from BSBolt.Utils.UtilityFunctions import get_external_paths, reverse_complement


class IndexOutput:
    """Class to output processed sequence index sequence and launch external index commands.

    """

    def __init__(self, genome_database: str = None, block_size: int = None):
        # format genome_database path
        bwa_path, _ = get_external_paths()
        self.genome_database = self.generate_genome_directory(genome_database)
        self.block_size = block_size
        self.bwa_path = bwa_path
        # set output object
        self.database_output = open(f'{self.genome_database}BSB_ref.fa', 'w')

    @staticmethod
    def generate_genome_directory(genome_database: str) -> str:
        """ Make directory if it doesn't exist, add / to output to ensure proper formatting
        Arguments:
            genome_database (str): output folder
        Returns:
             genome_database (str): formatted output path
        """
        if not os.path.isdir(genome_database):
            os.makedirs(genome_database, exist_ok=False)
        if not genome_database.endswith('/'):
            genome_database = f'{genome_database}/'
        return genome_database

    def write_contig_sequence(self, contig_id: str, contig_sequence: str):
        """ Writes formatted DNA sequence. Write possible outputs for Watson and Crick strands.
        Arguments:
            contig_id (str): contig label
            contig_sequence (str): str of DNA sequence
        """
        # write forward bisulfite converted sequence
        self.database_output.write(f'>{contig_id}\n')
        self.database_output.write(f'{contig_sequence}\n')

    def build_index(self):
        """Launch external commands for 4 processed reference files and collect external stdout for log file
        """
        # format output and input
        ref_file = f'{self.genome_database}BSB_ref.fa'
        # collect external command
        indx_command = [f'{self.bwa_path}', 'index', '-a', 'bwtsw', '-b', f'{self.block_size}', ref_file]
        # run external command
        subprocess.run(args=indx_command)

    def output_contig_sequence(self, contig_id: str, contig_sequence: Union[str, Dict[str, int]]):
        """Outputs serialized version of contig sequence
            Arguments:
                contig_id (str): contig label
                contig_sequence (object): str of DNA sequence
        """
        with open(f'{self.genome_database}{contig_id}.pkl', 'wb') as contig:
            return pickle.dump(contig_sequence, contig)

    def output_mappable_regions(self, mappable_regions: List[str]):
        """Outputs mappable regions
        Arguments:
            mappable_regions (list): list of bed formatted strings
                """
        with gzip.open(f'{self.genome_database}mappable_regions.txt.gz', 'wb') as mappable_regions_output:
            for line in mappable_regions:
                mappable_regions_output.write(line.encode('UTF-8'))



import gzip
import io
from typing import Dict, List, TextIO, Union
import numpy as np
from BSBolt.Matrix.SiteCounter import CGmapSiteCollector
from BSBolt.Matrix.SiteAggregator import CGmapSiteAggregator


def propagate_error(error):
    raise error


class AggregateMatrix:
    """
    Aggregate CGMap files into combined methylation matrix. CGmap files are first iterated through to count sites and
    then iterated to through to retrieve values. While slower this prevents the construction of large sparse matrices.
    Assembly is multi-threaded to improve performance.

    Params:

   * *file_list (list)*: list of file CGmap files
   * *sample_list (list)*: if passed, sample labels for CGmaps files,
                           else labels taken from sample names
   * *min_site_coverage (int)*: minimum read coverage for a CpG site to be considered for matrix, [10]
   * *site_proportion_threshold (float)*: proportion of samples that must have valid non-null
                                         methylation calls for a site to be included in matrix, [0.9]
   * *output_path (str)*: path to output file
   * *cg_only (bool)*: consider all cytosines or only CpG sites, [False]
   * *verbose (bool)*: verbose matrix assembly, [False]
   * *threads (int)*: threads available for matrix aggregation, [1]

    Attributes:

    * *self.sample_list (list)*: list of samples as they are ordered in the methylation matrix
    * *self.matrix_sites (tuple)*: tuple of ordered sites appearing in methylation matrix, only set if no output path
                                   is provided
    * *self.meth_matrix (np.array)*: array of methylation values (rows) by sample (columns), only set if not output path
                                    is provided

    Usage:

    ```python
    matrix = AggregateMatrix(**kwargs)
    matrix.aggregate_matrix()

    # access samples
    matrix.sample_list
    # access site list
    matrix.matrix_sites
    # access methylation matrix
    matrix.meth_matrix
    """

    def __init__(self, file_list: List[str] = None, sample_list: List[str] = None, min_site_coverage: int = 10,
                 site_proportion_threshold: float = 0.9, output_path: str = None, cg_only: bool = False,
                 verbose: bool = True, threads: int = 1, count_matrix: bool = False):
        self.file_list = file_list
        self.sample_list = sample_list
        if not self.sample_list:
            self.sample_list = []
            for file in self.file_list:
                self.sample_list.append(file.replace('\n', '').split('/')[-1])
        self.min_coverage = min_site_coverage
        self.verbose = verbose
        self.site_proportion_threshold = site_proportion_threshold
        self.output_path = output_path
        self.cg_only = cg_only
        self.threads = threads
        self.count_matrix = count_matrix
        self.meth_matrix = None
        self.matrix_sites = None

    def aggregate_matrix(self):
        """Iterate through passed CGmap files and aggregate matrix"""
        matrix_sites = self.collect_matrix_sites()
        meth_matrix = self.assemble_matrix(matrix_sites)
        if self.output_path:
            self.output_matrix(meth_matrix, matrix_sites)
        else:
            self.meth_matrix = meth_matrix
            self.matrix_sites = matrix_sites

    def collect_matrix_sites(self) -> Dict[str, int]:
        """Iterate through individual files to get consensus site counts"""
        site_collector = CGmapSiteCollector(cgmap_files=self.file_list,
                                            min_site_coverage=self.min_coverage,
                                            cg_only=self.cg_only,
                                            verbose=self.verbose,
                                            threads=self.threads)
        site_collector.collect_consensus_sites()
        matrix_sites = []
        site_count_threshold = int(self.site_proportion_threshold * len(self.sample_list))
        for site, count in site_collector.observed_site_count.items():
            if count >= site_count_threshold:
                matrix_sites.append(site)
        matrix_sites.sort(key=lambda x: int(x.split(':')[1]))
        matrix_sites.sort(key=lambda x: x.split(':')[0])
        return {site: count for count, site in enumerate(matrix_sites)}

    def assemble_matrix(self, matrix_sites: Dict[str, int]) -> np.ndarray:
        """Append sites to site list"""
        site_aggregator = CGmapSiteAggregator(cgmap_files=self.file_list,
                                              min_site_coverage=self.min_coverage,
                                              verbose=self.verbose,
                                              threads=self.threads,
                                              count_matrix=self.count_matrix)
        site_aggregator.assemble_matrix(matrix_sites=matrix_sites)
        return site_aggregator.meth_matrix

    def get_output_object(self) -> Union[TextIO, io.BufferedWriter]:
        if self.output_path.endswith('gz'):
            return io.BufferedWriter(gzip.open(self.output_path, 'wb'))
        return open(self.output_path, 'w')

    def output_matrix(self, meth_matrix: np.ndarray, matrix_sites: Dict[str, int]):
        """Output sorted aggregated matrix"""
        out = self.get_output_object()
        with out as matrix:
            sample_labels = '\t'.join([str(sample) for sample in self.sample_list])
            if self.count_matrix:
                sample_labels = '\t'.join([f'{sample}_meth_cytosine\t{sample}_total_cytosine'
                                          for sample in self.sample_list])
            matrix.write(f'Site\t{sample_labels}\n')
            for site, site_values in zip(matrix_sites, meth_matrix):
                if self.count_matrix:
                    meth_values = '\t'.join([f'{value}' for value in site_values])
                else:
                    meth_values = '\t'.join([f'{value:.6f}' for value in site_values])
                matrix.write(f'{site}\t{meth_values}\n')



#! /usr/env python3

import gzip
import io
from typing import Tuple, Union


class OpenFasta:
    """ Simple class to simplify iterating through fasta
    ------------------------------------------------------------------------------------
    KeywordArguments:
        fasta: str = path to fastq
    """

    def __init__(self, fasta: str = None):
        if fasta.endswith(".gz"):
            self.f = io.BufferedReader(gzip.open(fasta, 'rb'))
        else:
            self.f = open(fasta, 'r')

    def __iter__(self) -> Tuple[bool, str]:
        with self.f as fasta:
            while True:
                line = fasta.readline()
                if not line:
                    break
                processed_line = self.process_line(line)
                if '>' in processed_line:
                    yield True, processed_line
                else:
                    yield False, processed_line

    @staticmethod
    def process_line(line: Union[str, bytes]) -> str:
        if isinstance(line, bytes):
            return line.decode('utf-8').replace('\n', '')
        else:
            return line.replace('\n', '')



import random
from typing import Dict, Union, Tuple
from tqdm import tqdm
from BSBolt.Simulate.SetCyotsineMethylation import SetCytosineMethylation
from BSBolt.Simulate.StreamSim import StreamSim
from BSBolt.Utils.UtilityFunctions import get_external_paths, reverse_complement


class SimulateMethylatedReads:
    """Bisulifite read simulation class. The class works as follows:

    1. WGSIM (forked and modified version) is called to simulate paired end Illumina reads
        - If run in single end mode, the number of read simulated is double to get desired coverage and the second read
          isn't process as a bisulfite read or output.
    2. Methylation values are set for all methylatable bases (Cytosine and Guanine relative to the reference)
        - Values can be set randomly or taken from a reference file (BSBolt simulation database or CGmap file)
    3. Reads are bisulfite converted and output

    Params:

       * *reference_file (str)*: path to reference fasta file.
       * *sim_output (str)*: output path
       * *sequencing_error (float)*: simulated sequencing error rate, [0.005]
       * *mutation_rate (float)*: simulated mutation error rate, [0.0010]
       * *mutation_indel_fraction (float)*: fraction of mutations that are INDELs, [0.15]
       * *indel_extension_probability (float)*: probability INDEL length will be extended, [0.15]
       * *random_seed (int)*: random seed for mutation and sequencing error generation, [-1]
       * *paired_end (bool)*: simulate paired end bisulfite sequencing data. [False]
       * *read_length (int)*: length of simulated reads, [100]
       * *read_depth (int)*: average read depth over simulated contigs, [20]
       * *undirectional (bool)*: simulate undirectional (PCR product of Watson and Crick strands), [False]
       * *methylation_reference (str)*: path to previously generated BSBolt reference directory
       * *cgmap (str)*: path to CGmap file to use as methylation reference
       * *ambiguous_base_cutoff (float)*: reference segments where the proportion of ambiguous bases, - or N, greater
                                          than threshold will be skipped, [0.05]
       * *haplotype_mode (bool)*: simulate only homozygous variants, [False]
       * *pe_fragment_size (int)*: maximum fragment size, [400]
       * *insert_deviation (int)*: standard deviation of simulated insert sizes, [25]
       * *mean_insert_size (int)*: mean insert size, [100]
       * *collect_ch_sites (bool)*: simulated and collect CH methylation sites, [True]
       * *collect_sim_stats (bool)*: output simulated bases for all collected methylatable bases, [False]
       * *verbose (bool)*: verbose output, [True]
       * *overwrite_db (bool)*: overwrite previously generated BSBolt simulated database, [False]

    Usage:
    ```python
    simulation = SimulateMethylatedReads(**kwargs)
    simulation.run_simulation()
    ```

     """

    def __init__(self, reference_file: str = None, sim_output: str = None,
                 sequencing_error: float = 0.005, mutation_rate: float = 0.0010, mutation_indel_fraction: float = 0.15,
                 indel_extension_probability: float = 0.15, random_seed: int = -1,
                 paired_end: bool = False, read_length: int = 100,
                 read_depth: int = 20, undirectional: bool = False, methylation_reference: str = None,
                 cgmap: str = None, ambiguous_base_cutoff: float = 0.05, haplotype_mode: bool = False,
                 pe_fragment_size: int = 400, insert_deviation: int = 25, mean_insert_size: int = 100,
                 collect_ch_sites: bool = True, collect_sim_stats: bool = False, verbose: bool = True,
                 overwrite_db: bool = False):
        _, wgsim_path = get_external_paths()
        self.sim_command = [wgsim_path, '-1', str(read_length), '-2', str(read_length),
                            '-e', str(sequencing_error), '-d', str(pe_fragment_size),
                            '-s', str(insert_deviation),
                            '-r', str(mutation_rate), '-R', str(mutation_indel_fraction),
                            '-X', str(indel_extension_probability), '-S', str(random_seed),
                            '-A', str(ambiguous_base_cutoff), '-I', str(mean_insert_size)]
        if haplotype_mode:
            self.sim_command.append('-h')
        self.sim_db = SetCytosineMethylation(reference_file=reference_file,
                                             sim_dir=sim_output,
                                             methylation_reference=methylation_reference,
                                             cgmap=cgmap,
                                             collect_ch_sites=collect_ch_sites,
                                             overwrite_db=overwrite_db)
        self.sim_output = sim_output
        self.paired_end = paired_end
        self.output_objects = self.get_output_objects
        self.undirectional = undirectional
        self.read_coverage = (read_length, read_depth)
        self.reference = self.sim_db.reference
        self.collect_sim_stats = collect_sim_stats
        self.tqdm_disabe = False if verbose else True
        self.current_contig = None
        self.contig_profile = None
        self.contig_values = {}
        self.variant_data = {}

    def run_simulation(self):
        """Simulated bisulfite sequencing reads"""
        genome_length = sum([len(seq) for seq in self.reference.values()])
        coverage_length = self.read_coverage[0] * 2 if self.paired_end else self.read_coverage[0]
        read_number = (genome_length / coverage_length) * self.read_coverage[1]
        print('Simulating Methylated Reads')
        self.sim_command.extend(['-N', str(int(read_number)), self.sim_db.reference_file])
        self.simulate_methylated_reads()
        print('Finished Simulation')

    def simulate_methylated_reads(self):
        """Read processing loop"""
        for variant_contig, sim_data in tqdm(StreamSim(paired_end=self.paired_end, sim_command=self.sim_command),
                                             desc='Simulating Bisulfite Converted Read Pairs',
                                             disable=self.tqdm_disabe):
            if variant_contig:
                self.output_reference()
                self.current_contig = variant_contig
                self.contig_profile = self.get_methylation_reference(self.current_contig, sim_data)
                self.variant_data = sim_data
            else:
                if sim_data[1]['chrom'] != self.current_contig:
                    self.output_reference()
                    self.current_contig = sim_data[1]['chrom']
                    self.contig_profile = self.get_methylation_reference(self.current_contig)
                self.process_read_group(sim_data)
        self.sim_db.sim_db.output_contig(self.contig_profile, self.current_contig)
        self.output_reference()
        for output in self.output_objects:
            output.close()

    def output_reference(self):
        """Output reference methylation values used for simulation"""
        self.sim_db.sim_db.output_contig(self.contig_profile, self.current_contig)
        if self.variant_data:
            self.sim_db.sim_db.output_contig(self.variant_data, self.current_contig, variant=True)
        if self.collect_sim_stats:
            self.sim_db.sim_db.output_contig(self.contig_values, self.current_contig, values=True)
            self.contig_values = {}

    def process_read_group(self, sim_data):
        """Set read methylation values, randomly assign reads to Watson or Crick strand"""
        # randomly select reference strand
        sub_pattern = ('C', 'T') if self.random_roll(0.5) else ('G', 'A')
        # set read methylation
        self.set_read_methylation(sim_data[1], sub_base=sub_pattern[0])
        self.set_read_methylation(sim_data[2], sub_base=sub_pattern[0])
        # in silico bisulfite conversion
        sim_data[1]['seq'] = sim_data[1]['seq'].replace(sub_pattern[0], sub_pattern[1]).upper()
        if self.paired_end:
            sim_data[2]['seq'] = sim_data[2]['seq'].replace(sub_pattern[0], sub_pattern[1]).upper()
        # switch subpattern randomly for output if undirectional
        ref_strand = 'W' if sub_pattern[0] == 'C' else 'C'
        if self.undirectional:
            sub_pattern = ('C', 'T') if self.random_roll(0.5) else ('G', 'A')
        self.output_sim_reads(sim_data, sub_pattern[0], ref_strand)

    def output_sim_reads(self, sim_data, sub_base, ref_strand):
        """Write simulated bisulfite reads"""
        # format reads
        reverse_read = 2
        if sub_base == 'G':
            reverse_read = 1
        conversion_1, conversion_2 = ('C2T', 'G2A') if sim_data[1]['sub_base'] == sub_base else ('G2A', 'C2T')
        sim_data[reverse_read]['seq'] = reverse_complement(sim_data[reverse_read]['seq'])
        sim_data[reverse_read]['qual'] = sim_data[reverse_read]['qual'][::-1]
        sim_data[reverse_read]['cigar'] = sim_data[reverse_read]['cigar'][::-1]
        read_label = f'@{sim_data[1]["read_id"]}_{sim_data[1]["chrom"]}/1'
        read_comment = f'+{sim_data[1]["chrom"]}:{sim_data[1]["start"]}:' \
                       f'{sim_data[1]["end"]}:{sim_data[1]["cigar"]}:{ref_strand}{conversion_1}'
        read = f'{read_label}\n{sim_data[1]["seq"]}\n{read_comment}\n{sim_data[1]["qual"]}\n'
        self.output_objects[0].write(read)
        if self.paired_end:
            read_label = f'@{sim_data[2]["read_id"]}_{sim_data[2]["chrom"]}/2'
            read_comment = f'+{sim_data[1]["chrom"]}:{sim_data[2]["start"]}:' \
                           f'{sim_data[2]["end"]}:{sim_data[2]["cigar"]}:{ref_strand}{conversion_2}'
            read = f'{read_label}\n{sim_data[2]["seq"]}\n{read_comment}\n{sim_data[2]["qual"]}\n'
            self.output_objects[1].write(read)

    def set_read_methylation(self, read, sub_base='C'):
        """Set methylation according to sim value, variants can be methylation but not sequencing errors"""
        ref_seq, seq, cigar = self.reference[read['chrom']][read['start']: read['end']], read['seq'], read['cigar']
        # start iterators
        assert len(seq) == len(cigar)
        start, end = read['start'], read['end']
        methyl_read, methyl_cigar = list(seq), list(cigar)
        methyl_base_info = read['c_base_info'] if sub_base == 'C' else read['g_base_info']
        for site in methyl_base_info.split(','):
            if not site:
                break
            seq_pos, offset = site.split('_')
            seq_pos, offset = int(seq_pos), int(offset)
            cigar_base = cigar[seq_pos]
            ref_pos = start + seq_pos + offset
            seq_base = seq[seq_pos]
            ref_base = ref_seq[seq_pos + offset]
            if cigar_base == 'M':
                # check for sequencing error, if error don't evaluate methylation status
                meth_base, meth_base_cigar = self.handle_match(seq_base, ref_base, ref_pos)
            elif cigar_base == 'X':
                # handle methylatable conversions to converted base ie c to t, or g to a
                if seq_base != sub_base:
                    if seq_base == 'T' and sub_base == 'C':
                        methyl_cigar[seq_pos] = 'V'
                    elif seq_base == 'A' and sub_base == 'G':
                        methyl_cigar[seq_pos] = 'V'
                    continue
                meth_pos = f'{ref_pos}_{ref_base}_{seq_base}'
                meth_base, meth_base_cigar = self.set_variant_methylation(meth_pos, seq_base)
            else:
                meth_pos = f'{ref_pos}_{cigar_base}'
                meth_base, meth_base_cigar = self.set_variant_methylation(meth_pos, seq_base, 'I')
            if meth_base:
                methyl_read[seq_pos] = meth_base
                methyl_cigar[seq_pos] = meth_base_cigar
        read['seq'] = ''.join(methyl_read)
        read['cigar'] = ''.join(methyl_cigar)
        read['sub_base'] = sub_base

    def set_variant_methylation(self, meth_pos, seq_base, variant_type='X') -> Tuple[str, str]:
        cigar_type = 'Z' if variant_type == 'X' else 'R'
        methyl_status, context = self.set_base_methylation(meth_pos)
        if methyl_status:
            return seq_base.lower(), cigar_type
        else:
            return seq_base, cigar_type.lower()

    def handle_match(self, seq_base, ref_base, pos):
        if seq_base != ref_base:
            return seq_base, 'E'
        else:
            methyl_status, context = self.set_base_methylation(pos)
            if methyl_status:
                return seq_base.lower(), 'C' if context else 'Y'
            else:
                return seq_base, 'c' if context else 'y'

    def set_base_methylation(self, methyl_position):
        """Perform random roll to set methylated bases. Update reference values if *self.collect_sim_stats=True*"""
        # nucleotide, methylation_level, context, 0, 0, 0
        try:
            methyl_info = self.contig_profile[str(methyl_position)]
        except KeyError:
            return False, 0
        methyl_status = self.random_roll(methyl_info[1])
        if self.collect_sim_stats:
            if f'{self.current_contig}:{methyl_position}' not in self.contig_values:
                self.contig_values[f'{self.current_contig}:{methyl_position}'] = [0, 0]
            if methyl_status:
                self.contig_values[f'{self.current_contig}:{methyl_position}'][0] += 1
            else:
                self.contig_values[f'{self.current_contig}:{methyl_position}'][1] += 1
        return methyl_status, methyl_info[2]

    def get_methylation_reference(self, contig: str, variant_data: Union[bool, Dict] = False) -> Dict:
        """ Set variant methylation if variant data provided and return methylation profile else
        return methylation profile

        Params:

        * *contig (str)*: contig id
        * *variant_data (dict)*: simulated variant information

        Returns:

        * *contig_profile (Dict[str, float])*: methylation reference values"""
        if variant_data:
            contig_profile = self.sim_db.get_contig_methylation(contig)
            self.sim_db.set_variant_methylation(variant_data, contig_profile, self.current_contig)
            return contig_profile
        else:
            return self.sim_db.get_contig_methylation(contig)

    @property
    def get_output_objects(self):
        """Return io object for fastq writing"""
        output_list = [open(f'{self.sim_output}_1.fq', 'w')]
        if self.paired_end:
            output_list.append(open(f'{self.sim_output}_2.fq', 'w'))
        return output_list

    @staticmethod
    def random_roll(proportion_positive: float = 0.1) -> bool:
        """Return true is random < proportion_positive, used to set individual read methylation"""
        return random.random() < proportion_positive



#! /usr/env python3

import gzip
import io
from typing import List, Union


class OpenCGmap:

    def __init__(self, cgmap: str = None):
        if cgmap.endswith('.gz'):
            self.f = io.BufferedReader(gzip.open(cgmap, 'rb'))
        else:
            self.f = open(cgmap, 'r')

    def __iter__(self):
        with self.f as cg:
            while True:
                line = cg.readline()
                if not line:
                    break
                yield self.process_line(line)

    @staticmethod
    def process_line(line: Union[str, bytes]) -> List[str]:
        if isinstance(line, bytes):
            return line.decode('utf-8').replace('\n', '').split('\t')
        else:
            return line.replace('\n', '').split('\t')



#! /usr/env python3

import gzip
import io
from typing import Tuple, Union


class OpenAln:
    """ Simple class to simplify iterating through .aln files.
    ------------------------------------------------------------------------------------
    input: path to fastq
    returns: fastq iteration object"""

    def __init__(self, aln: str = None):
        if aln.endswith(".gz"):
            self.f = io.BufferedReader(gzip.open(aln, 'rb'))
        else:
            self.f = open(aln, 'r')

    def __iter__(self) -> Tuple[str, str, str]:
        with self.f as aln:
            aln_start = False
            while True:
                line1 = aln.readline()
                if not line1:
                    break
                if not aln_start:
                    if '##Header End' in line1:
                        aln_start = True
                    continue
                line1 = self.process_line(line1)
                line2 = self.process_line(aln.readline())
                line3 = self.process_line(aln.readline())
                yield line1, line2, line3

    @staticmethod
    def process_line(line: Union[str, bytes]) -> str:
        if isinstance(line, bytes):
            return line.decode('utf-8').replace('\n', '')
        else:
            return line.replace('\n', '')



#! /user/bin/env python3

from typing import List, Tuple


def chrom_site_split(chrom_site: str) -> Tuple[str, int]:
    site_split = chrom_site.split(':')
    assert len(site_split) == 2, 'Site must be formatted as chromosome:pos'
    try:
        position = int(site_split[1])
    except ValueError as e:
        print('Genome position must be numeric')
        raise e
    return site_split[0], position


class GenomeImputationWindows:
    """Calculates overlapping windows for imputation, assumes sites are sorted by coordinate
        Keyword Arguments:
            site_labels (list): list of genomic sites, chr:pos
            imputation_window_size (int): size of pairwise distance window
        Attributes:
            self.imputation_ws (int): size of window to impute values (middle third of window)
            self.imputation_boundary (int): size of window step size when getting overlapping windows
            self.site_dictionary (dict): sliding window stats
            self.window_count (int): increasing count of imputation windows, doesn't include global window
            self.windows (list): list of site indices for windows
            self.site_window_dict (dict): site hashed to imputation window
            self.rolling_windows (dict): site for rolling windows
        """

    def __init__(self, site_labels: List[str] = None, imputation_window_size: int = 3000000):
        self.site_labels = site_labels
        self.imputation_ws = int(imputation_window_size / 3) * 3
        self.imputation_boundary = int(imputation_window_size / 3)
        self.site_dictionary = None
        self.boundary_trigger = False
        self.window_count = 0
        self.windows = []
        self.site_window_dict = {}
        self.rolling_windows = {'window1': [], 'window2': [], 'window3': []}
        self.run()

    def run(self):
        self.set_initial_state()
        self.get_windows()

    def set_initial_state(self):
        """Set imputation starting point"""
        chrom, site = chrom_site_split(self.site_labels[0])
        self.site_dictionary = {'Chromosome': chrom, 'SitePosition': site,
                                'Window1Boundary': site + self.imputation_boundary,
                                'Window2Boundary': site + 2 * self.imputation_boundary,
                                'Window3Boundary': site + self.imputation_ws}

    def site_window(self, site):
        position = '%s:%s' % (self.site_dictionary['Chromosome'], str(site))
        if site <= self.site_dictionary['Window1Boundary']:
            self.rolling_windows['window1'].append(position)
            self.site_window_dict[position] = self.window_count
        elif self.site_dictionary['Window1Boundary'] < site <= self.site_dictionary['Window2Boundary']:
            self.rolling_windows['window1'].append(position)
            self.rolling_windows['window2'].append(position)
            self.site_window_dict[position] = self.window_count
        elif self.site_dictionary['Window2Boundary'] < site <= self.site_dictionary['Window3Boundary']:
            self.rolling_windows['window1'].append(position)
            self.rolling_windows['window2'].append(position)
            self.rolling_windows['window3'].append(position)
            self.site_window_dict[position] = self.window_count + 1
        elif site > self.site_dictionary['Window3Boundary']:
            self.boundary_trigger = True

    def advance_windows(self):
        self.window_count += 1
        self.windows.append(self.rolling_windows['window1'])
        self.rolling_windows['window1'] = list(self.rolling_windows['window2'])
        self.rolling_windows['window2'] = list(self.rolling_windows['window3'])
        self.rolling_windows['window3'] = []
        self.site_dictionary['Window1Boundary'] = int(self.site_dictionary['Window2Boundary'])
        self.site_dictionary['Window2Boundary'] = int(self.site_dictionary['Window3Boundary'])
        self.site_dictionary['Window3Boundary'] = int(self.site_dictionary['Window3Boundary']
                                                      + self.imputation_boundary)

    def reset_windows(self, chrom, site):
        """Save current window and reset window at new position"""
        self.windows.append(self.rolling_windows['window1'])
        self.window_count += 1
        self.site_dictionary = {'Chromosome': chrom, 'SitePosition': site,
                                'Window1Boundary': site + self.imputation_boundary,
                                'Window2Boundary': site + 2 * self.imputation_boundary,
                                'Window3Boundary': site + self.imputation_ws}
        self.rolling_windows = {'window1': [], 'window2': [], 'window3': []}
        position = '%s:%s' % (chrom, str(site))
        self.rolling_windows['window1'].append(position)
        self.site_window_dict[position] = self.window_count

    def get_windows(self):
        """Rules for window extension and window reset"""
        for position in self.site_labels:
            chrom, site = chrom_site_split(position)
            if chrom != self.site_dictionary['Chromosome']:
                self.reset_windows(chrom, site)
            else:
                # raise assertion error if file not sorted
                assert site >= self.site_dictionary['SitePosition'], f'File must be position sorted; ' \
                                                                     f'{chrom}:{site} < ' \
                                                                     f'{chrom}:{self.site_dictionary["SitePosition"]}'
                self.site_window(site)
                if self.boundary_trigger:
                    # advance windows
                    self.advance_windows()
                    self.boundary_trigger = False
                    self.site_window(site)
                    if self.boundary_trigger:
                        self.reset_windows(chrom, site)
                        self.boundary_trigger = False
        self.windows.append(self.rolling_windows['window1'])
        self.windows.append(self.rolling_windows['window2'])



#! /usr/bin/env python3

from typing import List
import numpy as np


def get_euclidean(array: np.ndarray = None, global_neighbors: np.ndarray = None) -> List[List[float]]:
    """ Pairwise euclidean distance
    ----------------------------------------
    input: numpy array with samples as rows
    output: pairwise euclidean distance array"""
    # split numpy array by columns
    assert isinstance(array, (np.ndarray, np.generic))
    samples = np.split(array, 1)
    # output to ensure correct matrix orientation
    # initialize empty list to hold pairwise distance values
    pairwise_distance = []
    # iterate through sample combinations
    for count, vector in enumerate(samples[0]):
        # add list for all sample pairwise comparison
        pairwise_distance.append([])
        for comparison_count, comparison_vector in enumerate(samples[0]):
            # only perform calculation for upper half of distance matrix to save time
            if comparison_count >= count:
                # stack sample values
                comparison_array = np.stack((vector, comparison_vector), axis=0)
                # drop any comparison value with NA for any sample, note this is only done pairwise between
                # samples to get best distance comparison estimate
                comparison_array = np.ma.compress_cols(np.ma.masked_invalid(comparison_array))
                try:
                    if len(comparison_array[0]) and len(comparison_array[1]):
                        pairwise_distance[count].append(np.linalg.norm(comparison_array[0] - comparison_array[1]) /
                                                        len(comparison_array[1]))
                    else:
                        raise IndexError
                # if all values are missing the missing neighbor is global
                except IndexError:
                    if isinstance(global_neighbors, (np.ndarray, np.generic)):
                        pairwise_distance[count].append(global_neighbors[count][comparison_count])
                    else:
                        pairwise_distance[count].append(np.nan)
            else:
                pairwise_distance[count].append(0)
    # add upper and lower matrices
    pairwise_upper = np.asarray(pairwise_distance)
    pairwise_lower = np.transpose(pairwise_upper)
    # return list to ensure compatibility with downstream tools
    pairwise_distance = (pairwise_upper + pairwise_lower).tolist()
    return pairwise_distance



import argparse

from BSBolt.Utils.UtilityFunctions import get_external_paths
from BSBolt.Utils.ParserHelpMessages import aggregate_help, alignment_help, impute_help, index_help, meth_help, sim_help

bt2_path, art_path = get_external_paths()

parser = argparse.ArgumentParser(description='BiSulfite Bolt, A bisulfite sequencing processing platform.',
                                 usage='BSBolt Module {Module Arguments}')

subparsers = parser.add_subparsers(description='BSBolt Modules, Please Invoke BSBolt Modules for Help\n '
                                               'Documentation at bsbolt.readthedocs.io',
                                   metavar='Index, Align, CallMethylation, AggregateMatrix, Simulate, Impute',
                                   dest='subparser_name')

align_parser = subparsers.add_parser('Align', help='Alignment Module', usage=alignment_help, add_help=False)
index_parser = subparsers.add_parser('Index', help='Index Generation Module', usage=index_help, add_help=False)
call_meth_parser = subparsers.add_parser('CallMethylation', help='Methylation Calling Module',
                                         usage=meth_help, add_help=False)
matrix_parser = subparsers.add_parser('AggregateMatrix', help='CGmap Matrix Aggregation Module', add_help=False,
                                      usage=aggregate_help)
sim_parser = subparsers.add_parser('Simulate', help='Read Simulation Module', add_help=False, usage=sim_help)
imputation_parser = subparsers.add_parser('Impute', help='kNN Imputation Module', add_help=False, usage=impute_help)
sort_parser = subparsers.add_parser('Sort', help='BAM Sort')
bam_index = subparsers.add_parser('BamIndex', help='BAM Index')
# Add Alignment Parser Commands

align_parser.add_argument('-F1', type=str, default=None, help='path to fastq 1', required=True)
align_parser.add_argument('-F2', type=str, default=None, help='path to fastq 2')
align_parser.add_argument('-UN', action="store_true", default=False,
                          help='library undirectional, ie. consider PCR products of bisulfite converted DNA',
                          required=False)
align_parser.add_argument('-O', type=str, default=None, help='output Prefix', required=True)
align_parser.add_argument('-DB', type=str, default=None, help='path to BSBolt database', required=True)
align_parser.add_argument('-CP', type=float, default=0.5, help='CH conversion proportion threshold [0.5]',
                          required=False)
align_parser.add_argument('-CT', type=int, default=5, help='number of CH sites needed to assess read conversion',
                          required=False)
align_parser.add_argument('-t', type=int, default=1, help='number of bwa threads [1]', required=False)
align_parser.add_argument('-k', type=int, default=19, help='minimum seed length [19]', required=False)
align_parser.add_argument('-w', type=int, default=100, help='band width for banded alignment [100]', required=False)
align_parser.add_argument('-d', type=int, default=100, help='off-diagonal X-dropoff [100]', required=False)
align_parser.add_argument('-r', type=float, default=1.5,
                          help='look for internal seeds inside a seed longer than {-k} * FLOAT [1.5]', required=False)
align_parser.add_argument('-y', type=int, default=20,
                          help='seed occurrence for the 3rd round seeding [20]', required=False)
align_parser.add_argument('-c', type=int, default=500,
                          help='skip seeds with more than INT occurrences [500]', required=False)
align_parser.add_argument('-D', type=float, default=0.50,
                          help='drop chains shorter than FLOAT fraction of the longest overlapping chain [0.50]',
                          required=False)
align_parser.add_argument('-W', type=int, default=0, help='discard a chain if seeded bases shorter than INT [0]',
                          required=False)
align_parser.add_argument('-m', type=int, default=50,
                          help='perform at most INT rounds of mate rescues for each read [50]',
                          required=False)
align_parser.add_argument('-S', action='store_true', default=False, help='skip mate rescue', required=False)
align_parser.add_argument('-P', action='store_true', default=False,
                          help='skip pairing; mate rescue performed unless -S also in use', required=False)
align_parser.add_argument('-A', type=int, default=1,
                          help='score for a sequence match, which scales options -TdBOELU unless overridden [1]',
                          required=False)
align_parser.add_argument('-B', type=int, default=4,
                          help='penalty for a mismatch [4]',
                          required=False)
align_parser.add_argument('-INDEL', type=lambda x: x.strip(), default='6,6',
                          help='gap open penalties for deletions and insertions [6,6]',
                          required=False)
align_parser.add_argument('-E', type=lambda x: x.strip(), default='1,1',
                          help='gap extension penalty; a gap of size k cost \'{-O} + {-E}*k\' [1,1]',
                          required=False)
align_parser.add_argument('-L', type=lambda x: x.strip(), default='30,30',
                          help='penalty for 5\'- and 3\'-end clipping [30,30]',
                          required=False)
align_parser.add_argument('-U', type=int, default='17',
                          help='penalty for an unpaired read pair [17]',
                          required=False)
align_parser.add_argument('-p', action='store_true', default=False,
                          help='smart pairing (ignoring in2.fq)',
                          required=False)
align_parser.add_argument('-R', type=str, default=None,
                          help='read group header line such as \'@RG\tID:foo\tSM:bar\' [null]',
                          required=False)
align_parser.add_argument('-H', type=str, default=None,
                          help='insert STR to header if it starts with @; or insert lines in FILE [null]',
                          required=False)
align_parser.add_argument('-j', action='store_true', default=False,
                          help='ignore ALT contigs',
                          required=False)
align_parser.add_argument('-T', type=int, default=80,
                          help='minimum score to output [80], set based on read length',
                          required=False)
align_parser.add_argument('-XA', type=lambda x: x.strip(), default='100,200',
                          help='if there are <INT hits with score >80 percent of the max score, '
                               'output all in XA [100,200]',
                          required=False)
align_parser.add_argument('-DR', type=float, default=0.95, help='drop ratio for alternative hits reported in XA tag, for best bisulifte alignment performance set at or above default [0.95]')
align_parser.add_argument('-M', action='store_true', default=False,
                          help='mark shorter split hits as secondary',
                          required=False)
align_parser.add_argument('-I', type=lambda x: x.strip(), default=None,
                          help='specify the mean, standard deviation (10 percent of the mean if absent), max '
                               '(4 sigma from the mean if absent) and min of the insert size distribution.  '
                               'FR orientation only. [inferred], Float,Float,Int,Int',
                          required=False)


# Add Index Parser Commands

index_parser.add_argument('-G', type=str, required=True,
                          help='Path to reference genome fasta file, fasta file should contain all contigs')
index_parser.add_argument('-DB', type=str, required=True,
                          help='Path to index directory, will create directory if folder does not exist')
index_parser.add_argument('-B', type=int, default=10000000, help='Block size for bwtsw algorithm, increasing '
                                                                 'block will speed up indexing and increase memory '
                                                                 'consumption')
index_parser.add_argument('-MR', type=str, default=None, help='Path to bed file of mappable regions. Index will be '
                                                              'built using using masked contig sequence')
index_parser.add_argument('-rrbs', action="store_true", default=False, help='Generate a Reduced Representative'
                                                                            ' Bisulfite Sequencing (RRBS) index')
index_parser.add_argument('-rrbs-cut-format', default='C-CGG',
                          help='Cut format to use for generation of RRBS database, '
                               'default= C-CGG (MSPI), input multiple enzymes as a '
                               'comma separate string, C-CGG,C-CGG,...')
index_parser.add_argument('-rrbs-lower', type=int, default=40, help='Lower bound fragment size to consider RRBS index'
                                                                    'generation, default = 40')
index_parser.add_argument('-rrbs-upper', type=int, default=500, help='Upper bound fragment size to consider RRBS index'
                                                                     'generation, default = 500')

# Add Methylation Calling Parser Arguments

call_meth_parser.add_argument('-I', type=str, required=True,
                              help='Input BAM, input file must be in BAM format with index file')
call_meth_parser.add_argument('-DB', type=str, required=True, help='Path to index directory')
call_meth_parser.add_argument('-O', type=str, required=True, help='Output prefix')
call_meth_parser.add_argument('-remove-ccgg', action="store_true", default=False,
                              help='Remove methylation calls in ccgg sites')
call_meth_parser.add_argument('-verbose', action="store_true", default=False, help='Verbose Output')
call_meth_parser.add_argument('-text', action="store_true", default=False,
                              help='Output plain text files, default=False')
call_meth_parser.add_argument('-ignore-ov', action="store_true", default=True,
                              help='Only consider higher quality base '
                                   'when paired end reads overlap')
call_meth_parser.add_argument('-max', type=int, default=8000, help='Max read depth to call methylation')
call_meth_parser.add_argument('-min', type=int, default=10,
                              help='Minimum read depth required to report methylation site')
call_meth_parser.add_argument('-t', type=int, default=1,
                              help='Number of threads to use when calling methylation values')
call_meth_parser.add_argument('-BQ', type=int, default=10, help='Minimum base quality for a base to considered for'
                                                               'methylation calling, default=0')
call_meth_parser.add_argument('-MQ', type=int, default=20, help='Minimum alignment quality for an alignment to be '
                                                                'considered for methylation calling, default=20')
call_meth_parser.add_argument('-CG', action="store_true", default=False, help='Only output CpG sites in CGmap file')
call_meth_parser.add_argument('-ATCG', action="store_true", default=False, help='Output ATCGmap file')
call_meth_parser.add_argument('-IO', action="store_true", default=False, help='Ignore orphans during methylation call')

# Add Matrix Aggregation Parser Args

matrix_parser.add_argument('-F', type=lambda file: [file_path for file_path in file.split(',')], required=True,
                           help='Comma separated list of CGmap file paths, or '
                                'path to text file with list of line separated '
                                'CGmap file paths')
matrix_parser.add_argument('-S', type=lambda sample_labels: [sample for sample in sample_labels.split(',')],
                           default=None,
                           help='Comma separated list of samples labels. '
                                'If sample labels are not provided sample labels '
                                'are extracted from CGmap file paths. '
                                'Can also pass path to txt for line separated sample '
                                'labels.')
matrix_parser.add_argument('-min-coverage', type=int, default=10, help='Minimum site read depth coverage for a '
                                                                       'site to be included in the aggregate matrix')
matrix_parser.add_argument('-min-sample', type=float, default=0.80,
                           help='Proportion of samples that must have a valid site '
                                '(above minimum coverage threshold), for a site to be'
                                'included in the aggregate matrix.')
matrix_parser.add_argument('-O', type=str, default=None, required=True, help='Aggregate matrix output path')
matrix_parser.add_argument('-CG', action="store_true", default=False, help='Only output CG sites')
matrix_parser.add_argument('-verbose', action="store_true", default=False, help='Verbose aggregation')
matrix_parser.add_argument('-t', type=int, default=1, help='Number of threads to use when assembling matrix')
matrix_parser.add_argument('-count', action='store_true', default=False,
                           help='Output a count matrix with count of methylated cytosines and '
                                'total observed cytosines, Sample_meth_cytosines   Sample_total_cytosines')


# Add Simulation Parser Args

sim_parser.add_argument('-G', type=str, required=True,
                        help='Path for reference genome fasta file, fasta file should contain all contigs')
sim_parser.add_argument('-O', type=str, required=True, help='Output prefix')
sim_parser.add_argument('-PE', default=False, action='store_true', help='Simulate Paired End Reads, default Single End')
sim_parser.add_argument('-RL', type=int, default=125, help='Simulated Read Length [125]')
sim_parser.add_argument('-RD', type=int, default=20, help='Simulated Read Depth [20]')
sim_parser.add_argument('-U', default=False, action='store_true',
                        help='Simulate Undirectional Reads, default=Directional')
sim_parser.add_argument('-CG', default=None, help='Path to CGmap file to generate simulation reference profile')
sim_parser.add_argument('-BR', default=None, help='Path to previously generated BSBolt methylation reference')
sim_parser.add_argument('-MR', type=float, default=0.005, help='Mutation rate [0.005]')
sim_parser.add_argument('-MI', type=float, default=0.20, help='Mutation indel fraction [0.20]')
sim_parser.add_argument('-ME', type=float, default=0.20, help='Mutation indel extension probability [0.20]')
sim_parser.add_argument('-RS', type=int, default=-1, help='Random seed for variant generation [-1]')
sim_parser.add_argument('-HA', default=False, action='store_true', help='Haplotype mode, homozygous variants only')
sim_parser.add_argument('-CH', default=True, action='store_false', help='Skip simulation of CH methylation, '
                                                                        'all CH sites unmethylated')
sim_parser.add_argument('-NS', default=True, action='store_false', help='By default observed methylation counts are '
                                                                        'saved, disable this behavior')
sim_parser.add_argument('-SE', type=float, default=0.001, help='Sequencing Error [0.001]')
sim_parser.add_argument('-NF', type=float, default=0.05, help='Cutoff threshold for amibiguous bases, simulated reads '
                                                              'with a proportion of ambiguous bases above this '
                                                              'threshold will not be output [0.05]')
sim_parser.add_argument('-FM', type=int, default=400, help='Max fragment size [400]')
sim_parser.add_argument('-IM', type=int, default=50, help='Insert length mean [50]')
sim_parser.add_argument('-SM', type=int, default=50, help='Insert length standard deviation [50]')
sim_parser.add_argument('-verbose', action='store_true', default=False, help='Verbose read simulation')
sim_parser.add_argument('-overwrite', action='store_true', default=False, help='Overwrite previously generated '
                                                                               'simulation database')


# Add Imputation Parser Args

imputation_parser.add_argument('-M', type=str, required=True, help='Path to BSBolt matrix file')
imputation_parser.add_argument('-B', type=int, default=0, help='Imputation sample batch size kNN imputation, by default'
                                                               ' the all of the samples will be processed as a single '
                                                               'batch')
imputation_parser.add_argument('-W', type=int, default=3000000, help='Sliding window size for imputation')
imputation_parser.add_argument('-k', type=int, default=5, help='Number of neighbors to use for imputation, default = 5')
imputation_parser.add_argument('-t', type=int, default=1, help='Number of threads available for imputation')
imputation_parser.add_argument('-verbose', action='store_true', default=False, help='Verbose output')
imputation_parser.add_argument('-O', type=str, default='', help='Output path for imputed matrix')
imputation_parser.add_argument('-R', action='store_true', default=False, help='Randomize batches')

# Add Sort Parser Args

sort_parser.add_argument('-O', type=str, required=True, help='Sorted bam output')
sort_parser.add_argument('-I', type=str, required=True, help='Input bam')

# add bam indexing args

bam_index.add_argument('-I', type=str, required=True, help='BAM input path')



import multiprocessing as mp
from typing import Dict, Deque, List
from BSBolt.Utils.CGmapIterator import OpenCGmap
from BSBolt.Utils.UtilityFunctions import propagate_error
import numpy as np
from tqdm import tqdm


class CGmapSiteAggregator:
    """Pull consensus methylation sites into a methylation matrix
       Keyword Arguments:
              cgmap_files (list): file paths
              min_site_coverage (int): minimum site coverage for a valid methylation call
              verbose (bool): tqdm output
              threads (int): threads available for aggregation
        """

    def __init__(self, cgmap_files: List[str] = None,
                 min_site_coverage: int = 1,
                 verbose: bool = False,
                 threads: int = 8,
                 count_matrix: bool = False):
        self.cgmap_files = cgmap_files
        self.min_site_coverage = min_site_coverage
        self.verbose = verbose
        self.threads = threads
        self.count_matrix = count_matrix
        self.meth_matrix = None

    def assemble_matrix(self, matrix_sites: Dict = None):
        # initialize collection_pool and manager queue
        matrix_pool = mp.Pool(processes=self.threads)
        manager = mp.Manager()
        matrix_queue = manager.Queue(maxsize=10)
        completed_samples = manager.list()
        # initialize empty array of zeros, fill with null values for safety
        if self.count_matrix:
            self.meth_matrix = np.zeros((len(matrix_sites), len(self.cgmap_files) * 2))
        else:
            self.meth_matrix = np.zeros((len(matrix_sites), len(self.cgmap_files)))
        for count, cgmap in enumerate(self.cgmap_files):
            matrix_pool.apply_async(self.collect_methylation_sites,
                                    kwds=dict(sample_pos=count, cgmap_path=cgmap, return_queue=matrix_queue,
                                              completed_samples=completed_samples, matrix_sites=matrix_sites),
                                    error_callback=propagate_error)
        matrix_pool.close()
        # if verbose initialize tqdm progress bar
        update_count = 0
        pbar = tqdm(total=len(self.cgmap_files), desc='Setting Methylation Sites') if self.verbose else None
        while True:
            if len(completed_samples) == len(self.cgmap_files) and matrix_queue.empty():
                break
            sample_pos, sample_values = matrix_queue.get(block=True)
            sample_span = sample_pos
            if self.count_matrix:
                sample_span = (sample_pos * 2, sample_pos * 2 + 1)
            self.meth_matrix[:, sample_span] = sample_values
            if pbar:
                update_number = len(completed_samples) - update_count
                pbar.update(update_number)
                update_count = len(completed_samples)
        if pbar:
            update_number = len(completed_samples) - update_count
            pbar.update(update_number)
            pbar.close()

    def collect_methylation_sites(self, sample_pos: int = None, cgmap_path: str = None, return_queue: Deque = None,
                                  completed_samples: List = None, matrix_sites: Dict = None):
        # initialize an array for all sites and fill with null values
        if self.count_matrix:
            sample_array = np.zeros((len(matrix_sites), 2))
        else:
            sample_array = np.zeros((len(matrix_sites)))
        sample_array.fill(np.nan)
        for line_info in OpenCGmap(cgmap_path):
            try:
                line_info[7]
            except IndexError:
                break
            else:
                if int(line_info[7]) >= self.min_site_coverage:
                    site_label = f'{line_info[0]}:{line_info[2]}'
                    if site_label in matrix_sites:
                        if self.count_matrix:
                            sample_array[matrix_sites[site_label], :] = np.array((int(line_info[6]), int(line_info[7])))
                        else:
                            sample_array[matrix_sites[site_label]] = float(line_info[5])
        return_queue.put((sample_pos, sample_array), block=True)
        completed_samples.append('done')



import os
from typing import List, Optional, Tuple
import numpy as np
from BSBolt.Utils.MatrixIterator import OpenMatrix


def get_bsb_matrix(bsb_matrix: str) -> Tuple[np.ndarray, List[str], List[str]]:
    """Import bsb matrix file
    Arguments:
        bsb_matrix (str): path to matrix file generated using BSBolt
    Returns:
        site_values (np.array): array of methylation matrix values
        site_order (list): list of methylation site identifiers
        sample_ids ([site_labels, [sample_labels]]: site label, list of sample labels
        """
    assert os.path.exists(bsb_matrix), f'{bsb_matrix} does not exist, please change path'
    site_values, site_order, sample_ids = [], [], None
    for line_label, line_values in OpenMatrix(bsb_matrix):
        if not sample_ids:
            sample_ids = line_label, line_values
        else:
            site_order.append(line_label)
            site_values.append(line_values)
    return np.asarray(site_values), site_order, sample_ids



import datetime
import os
import time
from BSBolt.Align.AlignReads import BisulfiteAlignmentAndProcessing
from BSBolt.CallMethylation.ProcessMethylationContigs import ProcessContigs
from BSBolt.Impute.kNN_Impute import ImputeMissingValues
from BSBolt.Index.RRBSIndex import RRBSBuild
from BSBolt.Index.WholeGenomeIndex import WholeGenomeBuild
from BSBolt.Matrix.MatrixAggregator import AggregateMatrix
from BSBolt.Simulate import SimulateMethylatedReads
from BSBolt.Utils.UtilityFunctions import index_bam, get_external_paths, sort_bam

bwa_path, wgsim_path = get_external_paths()


def launch_index(arguments):
    if arguments.rrbs:
        print(f'Generating RRBS Database at {arguments.DB}: '
              f'lower bound {arguments.rrbs_lower}, upper bound {arguments.rrbs_upper}: '
              f'Cut Format {arguments.rrbs_cut_format}')
        index = RRBSBuild(reference_file=arguments.G,
                          genome_database=arguments.DB,
                          cut_format=arguments.rrbs_cut_format,
                          lower_bound=arguments.rrbs_lower,
                          upper_bound=arguments.rrbs_upper,
                          block_size=arguments.B)
        index.generate_rrbs_database()
    else:
        print(f'Generating WGBS Database at {arguments.DB}')
        index = WholeGenomeBuild(reference_file=arguments.G,
                                 genome_database=arguments.DB,
                                 mappable_regions=arguments.MR,
                                 block_size=arguments.B)
        index.generate_bsb_database()


def align_bisulfite(bwa_cmd, output_path):
    start = time.time()
    print(' '.join(bwa_cmd))
    bs_alignment = BisulfiteAlignmentAndProcessing(bwa_cmd, output_path)
    bs_alignment.align_reads()
    alignment_time = datetime.timedelta(seconds=round(time.time() - start))
    print(f'Alignment Complete: Time {alignment_time}')
    print('------------------------------')
    mapping_stats = process_mapping_statistics(bs_alignment.mapping_statistics)
    print(mapping_stats)


def process_mapping_statistics(mapping_dict):
    dict(TotalReads=0, TotalAlignments=0, BSAmbiguous=0, C_C2T=0, C_G2A=0,
         W_C2T=0, W_G2A=0, Unaligned=0)
    processed_list = []
    mappability = (mapping_dict['TotalAlignments'] - mapping_dict['Unaligned']) / mapping_dict['TotalAlignments']
    processed_list.append(f'Total Reads: {mapping_dict["TotalReads"]}')
    processed_list.append(f'Mappability: {mappability * 100:.3f} %')
    processed_list.append('------------------------------')
    processed_list.append(f'Reads Mapped to Watson_C2T: {mapping_dict["W_C2T"]}')
    processed_list.append(f'Reads Mapped to Crick_C2T: {mapping_dict["C_C2T"]}')
    processed_list.append(f'Reads Mapped to Watson_G2A: {mapping_dict["W_G2A"]}')
    processed_list.append(f'Reads Mapped to Crick_G2A: {mapping_dict["C_G2A"]}')
    processed_list.append('------------------------------')
    processed_list.append(f'Unmapped Reads (Single / Paired Ends): {mapping_dict["Unaligned"]}')
    processed_list.append(f'Bisulfite Ambiguous: {mapping_dict["BSAmbiguous"]}')
    return '\n'.join(processed_list)


def launch_alignment(arguments):
    bsb_command_dict = {arg[0]: str(arg[1]) for arg in arguments._get_kwargs()}
    bwa_cmd = [bwa_path, 'mem', '-Y']
    if bsb_command_dict['UN'] == 'True':
        bwa_cmd.extend(['-z'])
    bool_args = ['M', 'S', 'j', 'p']
    for arg in bool_args:
        if bsb_command_dict[arg] == 'True':
            bwa_cmd.append(f'-{arg}')
    default_args = ['A', 'B', 'D', 'E', 'L', 'T', 'U', 'W', 'c', 'd', 'k', 'm', 'r', 't', 'w', 'y']
    for arg in default_args:
        bwa_cmd.extend([f'-{arg}', bsb_command_dict[arg]])
    if bsb_command_dict['H'] != 'None':
        bwa_cmd.extend([f'-H', bsb_command_dict['H']])
    if bsb_command_dict['I'] != 'None':
        bwa_cmd.extend([f'-I', bsb_command_dict['I']])
    if bsb_command_dict['INDEL']:
        bwa_cmd.extend([f'-O', bsb_command_dict['INDEL']])
    if bsb_command_dict['XA']:
        bwa_cmd.extend([f'-h', bsb_command_dict['XA']])
    bwa_cmd.extend(['-l', bsb_command_dict['CP']])
    bwa_cmd.extend(['-n', bsb_command_dict['CT']])
    bwa_cmd.extend(['-Z', bsb_command_dict['DR']])
    database = bsb_command_dict['DB']
    if not database.endswith('.fa'):
        if not database.endswith('/'):
            database = f'{database}/BSB_ref.fa'
        else:
            database = f'{database}BSB_ref.fa'
        assert os.path.exists(database), f'-DB {arguments.DB} does not exist, please index genome'
        assert os.path.exists(f'{database}.opac'), f'-DB {arguments.DB} not complete, please re-index genome'
    bwa_cmd.append(database)
    bwa_cmd.append(bsb_command_dict['F1'])
    assert os.path.exists(arguments.F1), f'-F1 {arguments.F1} does not exits, please check path'
    if bsb_command_dict['F2'] != 'None':
        bwa_cmd.append(bsb_command_dict['F2'])
        assert os.path.exists(arguments.F2), f'-F2 {arguments.F2} does not exits, please check path'
    align_bisulfite(bwa_cmd, arguments.O)


def launch_sort_bam(arguments):
    sort_bam(bam_output=arguments.O, bam_input=arguments.I)


def launch_index_bam(arguments):
    index_bam(bam_input=arguments.I)


def launch_methylation_call(arguments):
    if arguments.CG and arguments.ATCG:
        assert False, 'Reporting only CG sites notes compatible with .ATCGmap output'
    methylation_call = ProcessContigs(input_file=arguments.I,
                                      genome_database=arguments.DB,
                                      output_prefix=arguments.O,
                                      remove_ccgg=arguments.remove_ccgg,
                                      text_output=arguments.text,
                                      min_read_depth=arguments.min,
                                      threads=arguments.t,
                                      verbose=arguments.verbose,
                                      min_base_quality=arguments.BQ,
                                      min_mapping_quality=arguments.MQ,
                                      cg_only=arguments.CG,
                                      ATCGmap=arguments.ATCG,
                                      ignore_orphans=arguments.IO)
    methylation_call.process_contigs()
    methylation_call.watch_pool()


def launch_matrix_aggregation(arguments):
    def get_sample_info(file_path):
        file_list = []
        with open(file_path, 'r') as file:
            for line in file:
                file_list.append(line.replace('\n', ''))
        return file_list

    if len(arguments.F) == 1:
        arguments.F = get_sample_info(arguments.F[0])
    if arguments.S:
        if len(arguments.S) == 1:
            arguments.S = get_sample_info(arguments.S[0])

    matrix = AggregateMatrix(file_list=arguments.F,
                             sample_list=arguments.S,
                             min_site_coverage=arguments.min_coverage,
                             site_proportion_threshold=arguments.min_sample,
                             output_path=arguments.O,
                             cg_only=arguments.CG,
                             verbose=arguments.verbose,
                             threads=arguments.t,
                             count_matrix=arguments.count)
    matrix.aggregate_matrix()


def launch_simulation(arguments):
    read_simulation = SimulateMethylatedReads(reference_file=arguments.G,
                                              sim_output=arguments.O,
                                              sequencing_error=arguments.SE, mutation_rate=arguments.MR,
                                              mutation_indel_fraction=arguments.MI,
                                              indel_extension_probability=arguments.ME, random_seed=arguments.RS,
                                              paired_end=arguments.PE, read_length=arguments.RL,
                                              read_depth=arguments.RD, undirectional=arguments.U,
                                              methylation_reference=arguments.BR,
                                              cgmap=arguments.CG, ambiguous_base_cutoff=arguments.NF,
                                              haplotype_mode=arguments.HA,
                                              pe_fragment_size=arguments.FM, insert_deviation=arguments.SM,
                                              mean_insert_size=arguments.IM,
                                              collect_ch_sites=arguments.CH, collect_sim_stats=arguments.NS,
                                              verbose=arguments.verbose,
                                              overwrite_db=arguments.overwrite)

    read_simulation.run_simulation()


def launch_imputation(arguments):
    output_path = arguments.O
    if not output_path:
        output_path = f'{arguments.M}_imputed.txt'
    impute = ImputeMissingValues(input_matrix_file=arguments.M, batch_size=arguments.B,
                                 imputation_window_size=arguments.W, k=arguments.k, threads=arguments.t,
                                 verbose=arguments.verbose, output_path=output_path, randomize_batch=arguments.R)
    impute.import_matrix()
    impute.impute_values()
    impute.output_imputed_matrix()


bsb_launch = {'Index': launch_index,
              'Align': launch_alignment,
              'CallMethylation': launch_methylation_call,
              'AggregateMatrix': launch_matrix_aggregation,
              'Simulate': launch_simulation,
              'Impute': launch_imputation,
              'Sort': launch_sort_bam,
              'BamIndex': launch_index_bam}



from BSBolt.CallMethylation.CallValues import CallMethylationValues
from BSBolt.CallMethylation.CallVector import CallMethylationVector
from BSBolt.CallMethylation.ProcessMethylationContigs import ProcessContigs



import subprocess
from typing import List
import pysam


class BisulfiteAlignmentAndProcessing:
    """ Read alignment as processing. Handles reads from BWA-MEM2 and outputs to BAM file. Also aggregates alignment
    stats.

    Params:

    * *alignment_commands (list)*: bwa alignment commands
    * *output (str)*: output prefix

    Attributes:

    * self.mapping_statistics (dict)*: alignment run statistics
    """

    def __init__(self,  alignment_commands: List[str], output=None):
        self.alignment_commands = alignment_commands
        self.output = output
        self.mapping_statistics = dict(TotalReads=0, TotalAlignments=0, BSAmbiguous=0, C_C2T=0, C_G2A=0,
                                       W_C2T=0, W_G2A=0, Unaligned=0)

    def align_reads(self):
        """ Launch bwa alignment. Pipe output to BAM file
        """
        alignment_run = subprocess.Popen(self.alignment_commands,
                                         stdout=subprocess.PIPE,
                                         stderr=subprocess.PIPE,
                                         universal_newlines=True)
        # if non sam file is returned print stderr
        try:
            infile = pysam.AlignmentFile(alignment_run.stdout, 'r')
        except ValueError as e:
            for line in iter(alignment_run.stderr.readline, ''):
                print(line)
            raise e
        sam_out = pysam.AlignmentFile(f'{self.output}.bam', 'wb', template=infile)
        for s in infile:
            sam_out.write(s)
        self.update_mapping_statistics(alignment_run.stderr)

    def update_mapping_statistics(self, log_pipe):
        """Mapping statistics are updated per read group based on the mapping status"""
        for line in iter(log_pipe.readline, ''):
            if line[0:7] == 'BSStat ':
                category, count = line.replace('BSStat ', '').split(': ')
                self.mapping_statistics[category] += int(count)
            else:
                if line:
                    print(line.strip())



#! /user/bin/env python3

import multiprocessing as mp
from typing import List
import numpy as np
from tqdm import tqdm
from BSBolt.Impute.Imputation.GenomeImputationWindows import GenomeImputationWindows
from BSBolt.Impute.Imputation.EuclideanDistance import get_euclidean


class ImputationError(Exception):
    """Error in imputation process"""
    pass


def imputation_process_error(error):
    """Raise if exception thrown in methylation calling process"""
    raise ImputationError(error)


class GenomeImputation:
    """kNN imputation wrapper. Handles calculating nearest neighbors and setting missing values.
       KeywordArguments:
        row_labels (list): CpG Sites structured as chrom:pos
        sample_labels (list): sample labels
        genomic_array (np.array): array of methylation values with np.NaN representing empty sites
        imputation_window_size (int): size of total distance window, values are imputed for the middle third
        k (int): number of neighbors to pull values from
        threads (int): number of threads to calculate pairwise distance
        verbose (bool): default=False, tqdm output
       Attributes
        self.row_labels (list): CpG Sites structured as chrom:pos
        self.sample_labels (list): sample labels
        self.genomic_array (np.array): array of methylation values with np.NaN representing empty sites
        self.imputation_window_size (int): size of total distance window, values are imputed for the middle third
        self.k (int): number of neighbors to pull values from
        self.threads (int): number of threads to calculate pairwise distance
        self.verbose (bool): default=False, tqdm output
        self.pool (mp.pool): multiprocessing kNN pool
        self.global_neighbors (dict): reference values to use for empty imputation windows
        self.neighbors_dicts (dict) imputation pairwise distance neighbors
        self.genomic_array_t (np.array): reference to transposed version of self.genomic_array
           """

    def __init__(self, row_labels: List[str] = None, sample_labels: List[str] = None,
                 genomic_array: np.ndarray = None, imputation_window_size: int = 3000000, k: int = 5,
                 threads: int = 4, verbose: bool = False):
        self.row_labels = row_labels
        self.sample_labels = sample_labels
        self.genomic_array = genomic_array
        assert isinstance(self.genomic_array, (np.ndarray, np.generic)), f'{self.genomic_array is None}'
        # check that matrix is properly oriented for imputation
        if genomic_array.shape[0] < genomic_array.shape[1]:
            raise RuntimeWarning(f'Number of Genomic Locations {genomic_array.shape[0]} < Number of Samples '
                                 f'{genomic_array.shape[1]}, Transpose Matrix if Rows Represent Samples')
        self.imputation_window_size = imputation_window_size
        self.k = k
        self.threads = threads
        self.verbose = verbose
        self.pool = None
        self.global_neighbors = None
        self.imputation_windows = None
        self.neighbor_dicts = None
        self.genomic_array_t = genomic_array.T

    def impute_windows(self):
        """Imputation run command"""
        self.get_global_neighbors()
        self.get_imputation_windows()
        self.get_window_values()
        self.watch_pool()
        self.set_missing_values()

    def get_global_neighbors(self):
        """Calculate global neighbors for sites without local information"""
        self.global_neighbors: List[List[float]] = get_euclidean(self.genomic_array_t)

    def get_imputation_windows(self):
        """Retrieve imputation windows based on input row labels"""
        self.imputation_windows = GenomeImputationWindows(site_labels=self.row_labels,
                                                          imputation_window_size=self.imputation_window_size)

    def get_window_values(self):
        """Impute pairwise distance for all imputation windows. Imputation windows are computed in parallel"""
        self.pool = mp.Pool(processes=self.threads)
        window_manager = mp.Manager()
        self.neighbor_dicts = window_manager.dict()
        for window_count, window in enumerate(self.imputation_windows.windows):
            if window:
                first_slice = self.row_labels.index(window[0])
                second_slice = self.row_labels.index(window[-1]) + 1
                window_values = np.copy(self.genomic_array[first_slice:second_slice, :])
                pairwise_distance_kwargs = dict(return_dict=self.neighbor_dicts,
                                                window_count=window_count,
                                                array=window_values,
                                                global_neighbors=self.global_neighbors,
                                                sample_labels=self.sample_labels,
                                                k=self.k)
                self.pool.apply_async(self.get_pairwise_distance, kwds=pairwise_distance_kwargs,
                                      error_callback=imputation_process_error)
            else:
                self.neighbor_dicts[window_count] = 'Empty Window'

    def watch_pool(self):
        """Watch pairwise distance operation and update status
        """
        completed_windows = None
        pbar = None
        if self.verbose:
            pbar = tqdm(total=len(self.imputation_windows.windows), desc='Calculating Window Pairwise Distance')
            completed_windows = len(self.neighbor_dicts)
            pbar.update(completed_windows)
        while len(self.neighbor_dicts) != len(self.imputation_windows.windows):
            if self.verbose:
                new_windows = len(self.neighbor_dicts) - completed_windows
                pbar.update(new_windows)
                completed_windows = len(self.neighbor_dicts)
        if self.verbose:
            new_windows = len(self.neighbor_dicts) - completed_windows
            pbar.update(new_windows)
            pbar.close()
        self.pool.close()

    @staticmethod
    def get_pairwise_distance(return_dict=None, window_count=None, global_neighbors=None,
                              array=None, sample_labels=None, k=None):
        """Method to calculate pairwise distance and return nearest neighbors. Distance matrix is discarded to
        reduce the memory overhead"""
        distance_matrix = get_euclidean(array.T, global_neighbors)
        local_neighbor_dict = {}
        for sample in zip(sample_labels, distance_matrix):
            sorted_sample_key = [a for b, a in sorted(zip(sample[1], sample_labels))]
            neighbors = [sample_labels.index(x) for x in sorted_sample_key[1:(k + 1)]]
            local_neighbor_dict[sample_labels.index(sample[0])] = neighbors
        return_dict[window_count] = local_neighbor_dict

    def set_missing_values(self):
        """Set imputed methylation values in genomic array and return results """
        # ensure all label are strings to set on label and not index
        imputed_values = []
        for label, row in tqdm(zip(self.row_labels, self.genomic_array), total=len(self.row_labels),
                               desc='Setting Imputed Values', disable=False if self.verbose else True):
            current_values = np.copy(row)
            null_methylation_values = np.argwhere(np.isnan(current_values))
            if len(null_methylation_values) > 0:
                null_site_neighbor_dict = self.neighbor_dicts[self.imputation_windows.site_window_dict[label]]
                for null_site in null_methylation_values:
                    neighbors = null_site_neighbor_dict[null_site[0]]
                    neighbor_average = round(np.nanmean(row[neighbors]), 5)
                    current_values[null_site[0]] = neighbor_average
            imputed_values.append(current_values)
        self.genomic_array = np.array(imputed_values)



#! /usr/env python3

import gzip
import io
from typing import List, Union


class OpenFastq:
    """ Simple class to simplify iterating through fastq files. The script yields a tuple for every four lines
    in a fastq file
    ------------------------------------------------------------------------------------
    KeywordArguments:
        fastq: str = path to fastq file
    """

    def __init__(self, fastq: str = None):
        if fastq.endswith(".gz"):
            self.f = io.BufferedReader(gzip.open(fastq, 'rb'))
        else:
            self.f = open(fastq, 'r')

    def __iter__(self) -> List[str]:
        with self.f as fastq:
            while True:
                line1 = fastq.readline()
                if not line1.strip():
                    break
                line1 = self.process_line(line1)
                line2 = self.process_line(fastq.readline())
                line3 = self.process_line(fastq.readline())
                line4 = self.process_line(fastq.readline())
                yield [line1, line2, line3, line4]

    @staticmethod
    def process_line(line: Union[bytes, str]):
        if isinstance(line, bytes):
            return line.decode('utf-8').replace('\n', '')
        else:
            return line.replace('\n', '')



from distutils.version import LooseVersion
import os
import sys
import pysam


def reverse_complement(sequence):
    """
    Params:

    * *sequence (str)*: DNA sequence, non ATGC nucleotide will be returned unaltered

    Returns:

    * *reversed_string.translate(_rc_trans) (str)*: reverse complement of input sequence
    """
    # reverse string
    reversed_string = sequence[::-1]
    # replace base with complement
    _rc_trans = str.maketrans('ACGTNacgtn', 'TGCANtgcan')
    return reversed_string.translate(_rc_trans)


def retrieve_iupac(nucleotide):
    """
    Params:

    * *nucleotide (str)*: single character

    Returns:

    * *iupac_tuple (tuple)*: tuple of strings with possible bases
    """
    iupac_key = {'R': ('A', 'G'), 'Y': ('C', 'T'), 'S': ('G', 'C'), 'W': ('A', 'T'), 'K': ('G', 'T'), 'M': ('A', 'C'),
                 'B': ('C', 'G', 'T'), 'D': ('A', 'G', 'T'), 'H': ('A', 'C', 'T'), 'V': ('A', 'C', 'G'),
                 'N': ('A', 'C', 'G', 'T')}
    try:
        iupac_tuple = iupac_key[nucleotide]
    except KeyError:
        iupac_tuple = tuple(nucleotide)
    return iupac_tuple


def import_package_check(package_name):
    try:
        package = __import__(package_name)
    except ModuleNotFoundError:
        print(f'{package_name} not found, please install {package_name}')
        print(f'pip3 install {package_name} --user')
        sys.exit()
    else:
        return package.__version__


def check_package_version():
    """Check third party packages to make sure version is greater than requirement"""
    packages = dict(pysam='0.15.2', tqdm='4.31.1', numpy='1.16.3')
    all_packages = True
    for package, required_version in packages.items():
        installed_version = import_package_check(package)
        if LooseVersion(installed_version) < LooseVersion(required_version):
            print(f'{package} {installed_version} < {package} {required_version}, please update {package}')
            print(f'pip3 install {package} --upgrade')
            all_packages = False
    return all_packages


def check_python_version():
    if sys.version_info < (3, 6):
        print('Python must be >= 3.6.0')
        raise OSError


def get_external_paths():
    """Get paths of dependencies. Print warning if setup.py not run and dependencies not compiled.

    Returns:
    * *bwa (str)*: path to bwa executable
    * *wgsim (str)*: path to wgsim executable
    """
    utility_directory = os.path.dirname(os.path.realpath(__file__))
    external_directory = '/'.join(utility_directory.split('/')[:-1]) + '/External/'
    bwa = f'{external_directory}BWA/bwa'
    wgsim = f'{external_directory}WGSIM/wgsim'
    if not os.path.exists(bwa) or not os.path.exists(wgsim):
        print(f'Must compile external dependencies\n python3 setup.py build')
    return bwa, wgsim


def propagate_error(error):
    raise error


def sort_bam(bam_output=None, bam_input=None):
    """ Sort bam file

    Params:
    * *bam_output (str)*: output path for sorted bam file
    * *bam_input (str)*: input bam file"""
    pysam.sort('-o', bam_output, bam_input)


def index_bam(bam_input=None):
    """
    Index bam file

    Params:
    * *bam_input (str)*: input bam files
    """
    pysam.index(bam_input)



import os
from setuptools import Extension, setup
from setuptools.command.develop import develop
from setuptools.command.build_py import build_py
import subprocess

with open("README.md", "r") as fh:
    long_description = fh.read()

"""
External dependencies compiled for linux using quay.io/pypa/manylinux2010_x86_64 image:
/opt/python/cpXX-cpXX/bin/python setup.py bdist_wheel
"""

bwa_ext = Extension('bwa', sources=['BSBolt/External/BWA/bwa'])
wgsim_ext = Extension('wgsim', sources=['BSBolt/External/WGSIM/wgsim'])


class BuildError(Exception):
    """Error in methylation calling process"""
    pass


def compile_dependency(compilation_command, cwd):
    comp = subprocess.Popen(compilation_command, stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            universal_newlines=True, cwd=cwd)
    comp.wait()
    if comp.returncode:
        print('Please compile with GCC >= 8.3.1 / zlib-devel >= 1.2.3')
        for line in iter(comp.stderr.readline, ''):
            print(line)
        raise BuildError
    for line in iter(comp.stdout.readline, ''):
        formatted_line = line.strip()
        print(formatted_line)


def make_external_dependencies():
    working_directory = os.path.dirname(os.path.realpath(__file__))
    bwa_directory = f'{working_directory}/BSBolt/External/BWA'
    wgsim_directory = f'{working_directory}/BSBolt/External/WGSIM'
    if not os.path.exists(f'{wgsim_directory}/wgsim'):
        print('Compiling wgsim')
        compile_dependency(['make'], wgsim_directory)
    if not os.path.exists(f'{bwa_directory}/bwa-mem2'):
        print('Compiling bwa-mem')
        compile_dependency(['make'], bwa_directory)


try:
    from wheel.bdist_wheel import bdist_wheel as _bdist_wheel

    class bdist_wheel(_bdist_wheel):
        def finalize_options(self):
            _bdist_wheel.finalize_options(self)
            self.root_is_pure = False
except ImportError:
    bdist_wheel = None


class DevelopCmd(develop):

    def run(self):
        make_external_dependencies()
        super().run()


class BuildCmd(build_py):

    def run(self):
        make_external_dependencies()
        self.data_files = [('BSBolt', 'BSBolt', 'build/lib/BSBolt', []),
                           ('BSBolt.Align', 'BSBolt/Align', 'build/lib/BSBolt/Align', []),
                           ('BSBolt.CallMethylation', 'BSBolt/CallMethylation', 'build/lib/BSBolt/CallMethylation', []),
                           ('BSBolt.External', 'BSBolt/External', 'build/lib/BSBolt/External',
                            ['WGSIM/wgsim', 'BWA/bwa']),
                           ('BSBolt.Impute', 'BSBolt/Impute', 'build/lib/BSBolt/Impute', []),
                           ('BSBolt.Impute.Imputation', 'BSBolt/Impute/Imputation',
                            'build/lib/BSBolt/Impute/Imputation', []),
                           ('BSBolt.Impute.Impute_Utils', 'BSBolt/Impute/Impute_Utils',
                            'build/lib/BSBolt/Impute/Impute_Utils', []), (
                           'BSBolt.Impute.Validation', 'BSBolt/Impute/Validation', 'build/lib/BSBolt/Impute/Validation',
                           []),
                           ('BSBolt.Index', 'BSBolt/Index', 'build/lib/BSBolt/Index', []),
                           ('BSBolt.Matrix', 'BSBolt/Matrix', 'build/lib/BSBolt/Matrix', []),
                           ('BSBolt.Simulate', 'BSBolt/Simulate', 'build/lib/BSBolt/Simulate', []),
                           ('BSBolt.Utils', 'BSBolt/Utils', 'build/lib/BSBolt/Utils', []),
                           ('BSBolt.Variant', 'BSBolt/Variant', 'build/lib/BSBolt/Variant', [])]
        super().run()


setup(name='BSBolt',
      version='1.1.1',
      description='Bisulfite Sequencing Processing Platform',
      long_description=long_description,
      long_description_content_type="text/markdown",
      url='https://github.com/NuttyLogic/BSBolt',
      project_urls={'Documentation': 'https://bsbolt.readthedocs.io/en/latest/'},
      author='Colin P. Farrell',
      author_email='colinpfarrell@gmail.com',
      license='MIT',
      packages=['BSBolt',
                'BSBolt.Align',
                'BSBolt.CallMethylation',
                'BSBolt.Impute',
                'BSBolt.Impute.Imputation',
                'BSBolt.Impute.Impute_Utils',
                'BSBolt.Impute.Validation',
                'BSBolt.Index',
                'BSBolt.Matrix',
                'BSBolt.Simulate',
                'BSBolt.Utils',
                'BSBolt.Variant'],
      classifiers=['Programming Language :: Python :: 3.6',
                   'Programming Language :: Python :: 3.7',
                   'Programming Language :: Python :: 3.8'],
      platforms=["Linux", "Mac OS-X", "Unix"],
      requires=['pysam', 'numpy', 'tqdm'],
      install_requires=['pysam>=0.15.4', 'numpy>=1.16.3', 'tqdm>=4.31.1'],
      entry_points={'console_scripts': ['BSBolt = BSBolt.__main__:launch_bsb']},
      python_requires='>=3.6',
      test_suite='tests',
      include_package_data=True,
      cmdclass={'develop': DevelopCmd,
                'build_py': BuildCmd,
                'bdist_wheel': bdist_wheel},
      zip_safe=False,
      extension_modules=[bwa_ext, wgsim_ext]
      )



import pickle
import os


class SimulationOutput:

    def __init__(self, sim_dir=None):
        self.sim_dir = sim_dir

    def generate_simulation_directory(self):
        simulation_directory = '/'.join(self.sim_dir.split('/')[0:-1])
        if simulation_directory:
            if not os.path.isdir(simulation_directory):
                os.makedirs(simulation_directory, exist_ok=False)

    def output_contig(self, contig_profile, contig_id, values=False, variant=False):
        if contig_id:
            contig_label = contig_id
            if values:
                contig_label = f'{contig_id}_values'
            elif variant:
                contig_label = f'{contig_id}_variants'
            with open(f'{self.sim_dir}.{contig_label}.pkl', 'wb') as contig_out:
                pickle.dump(contig_profile, contig_out)

    def load_contig(self, contig_id, values=False):
        contig_label = contig_id
        if values:
            contig_label = f'{contig_id}_values'
        try:
            with open(f'{self.sim_dir}.{contig_label}.pkl', 'rb') as contig_out:
                contig_profile = pickle.load(contig_out)
        except FileNotFoundError:
            return None
        else:
            return contig_profile




#! /usr/env python3

import gzip
import io
from typing import List, Tuple, Union
import numpy as np


class OpenMatrix:
    """ Simple class to simple class to iterate through BSBolt matrix file
    ------------------------------------------------------------------------------------
    input: path to matrix file
    returns: matrix iteration object"""

    def __init__(self, matrix: str = None):
        self.header = False
        if matrix.endswith(".gz"):
            self.f = io.BufferedReader(gzip.open(matrix, 'rb'))
        else:
            self.f = open(matrix, 'r')

    def __iter__(self):
        with self.f as matrix:
            while True:
                line = matrix.readline()
                if not line.strip():
                    break
                line = self.process_line(line)
                yield line

    def process_line(self, line) -> Tuple[str, Union[List[str], np.ndarray]]:
        converted_line = self.line_conversion(line)
        if not self.header:
            self.header = True
            return converted_line[0], converted_line[1:]
        return converted_line[0], np.asarray([self.convert_to_float(value) for value in converted_line[1:]])

    @staticmethod
    def line_conversion(line) -> List[str]:
        if isinstance(line, bytes):
            return line.decode('utf-8').replace('\n', '').split('\t')
        else:
            return line.replace('\n', '').split('\t')

    @staticmethod
    def convert_to_float(value: str) -> Union[float, None]:
        try:
            return float(value)
        except ValueError:
            return np.nan



from typing import Dict, List, Tuple

from BSBolt.Index.IndexOutput import IndexOutput
from BSBolt.Utils.FastaIterator import OpenFasta


class WholeGenomeBuild:
    """Class to build whole genome bisulfite bwa index.

    Params:

    * *reference_file (str)*: path to reference file in fasta format
    * *genome_database (str)*: path to genome database output
    * *mappable_regions (str)*: path to bed file of mappable regions for masked alignment index building
    * *block_size (int)*: bwa indexing block size, increases indexing speed but increases memory consumption

    Usage:
    ```python
    index = WholeGenomeBuild(**kwargs)
    index.generate_bsb_database()
    ```
    """

    def __init__(self, reference_file: str = None, genome_database: str = None,
                 mappable_regions: str = None, block_size: int = None):
        self.reference_file = OpenFasta(fasta=reference_file)
        self.index_output = IndexOutput(genome_database=genome_database, block_size=block_size)
        self.mappable_regions = None
        if mappable_regions:
            self.mappable_regions = self.get_mappable_regions(mappable_regions)
        self.contig_size_dict = {}

    def generate_bsb_database(self):
        """ Wrapper for class functions to process and build mapping indices.
        """
        contig_sequence = []
        contig_id = None
        # return true, line if '>' in fasta line
        for is_contig_label, sequence in self.reference_file:
            if is_contig_label:
                if contig_id:
                    # join contig sequence
                    contig_str = ''.join(contig_sequence)
                    if self.mappable_regions:
                        contig_str = self.mask_contig(contig_id, contig_str)
                    # serialize contig file
                    self.index_output.output_contig_sequence(contig_id, contig_str)
                    # output contig reference
                    self.index_output.write_contig_sequence(contig_id, contig_str)
                    # get contig length
                    self.contig_size_dict[contig_id] = len(contig_str)
                    contig_sequence = []
                # reset contig_id
                contig_id = sequence.replace('>', '').split()[0]
            else:
                contig_sequence.append(sequence)
        # process remaining contig sequence
        contig_str = ''.join(contig_sequence)
        contig_len = len(contig_str)
        if self.mappable_regions:
            contig_str = self.mask_contig(contig_id, contig_str)
            assert contig_len == len(contig_str), "Indexing error masked contig larger than reference"
        self.index_output.output_contig_sequence(contig_id, contig_str)
        self.index_output.write_contig_sequence(contig_id, contig_str)
        self.index_output.database_output.close()
        self.contig_size_dict[contig_id] = len(contig_str)
        self.index_output.output_contig_sequence('genome_index', self.contig_size_dict)
        # launch external commands
        self.index_output.build_index()

    @staticmethod
    def get_mappable_regions(bed_file: str) -> Dict[str, List[Tuple[int, int]]]:
        """Get mappable regions if masking with bed file

        Params:

        * *bed_file (str)*: path to bed file of mappable regions

        Returns:

        * *mappable_regions (list)*: sorted list of mappable regions
        """
        mappable_regions = {}
        with open(bed_file, 'r') as regions:
            for bed_line in regions:
                if bed_line:
                    chrom, start, end = bed_line.replace('\n', '').split('\t')[0:3]
                    start, end = int(start), int(end)
                    if chrom in mappable_regions:
                        mappable_regions[chrom].append((start - 2, end + 2))
                    else:
                        mappable_regions[chrom] = [(start - 2, end + 2)]
        for region_list in mappable_regions.values():
            region_list.sort(key=lambda x: x[0])
        return mappable_regions

    def mask_contig(self, contig_id: str, contig_str: str) -> str:
        """Mask contig sequence outside mappable regions

        Params:

        * *contig_id (str)* contig label
        * *contig_str (str)*: contig sequence

        Returns:

        * *contig_str (str)*: masked sequence
        """
        if contig_id in self.mappable_regions:
            contig_mappable_regions = iter(self.mappable_regions[contig_id])
            start, end = next(contig_mappable_regions)
            masked_contig = []
            for position, bp in enumerate(contig_str):
                if position > end:
                    try:
                        start, end = next(contig_mappable_regions)
                    except StopIteration:
                        pass
                if start <= position <= end:
                    masked_contig.append(bp)
                else:
                    masked_contig.append('-')
            return ''.join(masked_contig)
        else:
            return '-' * len(contig_str)
